using System.IO;
using System.Linq;

using Raven.Database.Config;
using Raven.Database.FileSystem.Synchronization.Rdc.Wrapper;
using Xunit;
using Xunit.Extensions;

namespace Raven.Tests.FileSystem
{
    public class StorageSignatureRepositoryTests : StorageTest
    {
        private InMemoryRavenConfiguration configuration;

        public StorageSignatureRepositoryTests()
        {
            configuration = new InMemoryRavenConfiguration();
            configuration.Initialize();
        }

        [Fact]
        public void Should_read_from_storage()
        {
            transactionalStorage.Batch(accessor =>
            {
                accessor.AddSignature("test", 1, stream => stream.Write(new byte[] { 3 }, 0, 1));
            });

            var tested = new StorageSignatureRepository(transactionalStorage, "test", configuration);
            Assert.Equal(3, tested.GetContentForReading("test.1.sig").ReadByte());
        }

        [Fact]
        public void Should_throw_FileNotFoundException_for_unknown_file()
        {
            transactionalStorage.Batch(accessor =>
            {
                accessor.AddSignature("test", 1, stream => stream.Write(new byte[] { 3 }, 0, 1));
            });

            var tested = new StorageSignatureRepository(transactionalStorage, "test", configuration);
            Assert.Throws(typeof(FileNotFoundException), () => tested.GetContentForReading("test.0.sig"));
        }

        [Fact]
        public void Should_get_SignatureInfo()
        {
            transactionalStorage.Batch(accessor =>
            {
                accessor.AddSignature("test", 1, stream => stream.Write(new byte[] { 3 }, 0, 1));
            });
            var tested = new StorageSignatureRepository(transactionalStorage, "test", configuration);
            var result = tested.GetByName("test.1.sig");
            Assert.Equal("test.1.sig", result.Name);
            Assert.Equal(1, result.Length);
        }

        [Theory]
        [InlineData("test.bin")]
        [InlineData("/directory/test.bin")]
        public void Should_assign_signature_to_proper_file(string fileName)
        {
            var tested = new StorageSignatureRepository(transactionalStorage, fileName, configuration);
            using(var sigContent = tested.CreateContent(fileName + ".0.sig"))
            {
                sigContent.WriteByte(3);
            }
            tested.Flush(new[] { SignatureInfo.Parse(fileName + ".0.sig") } );

            var result = tested.GetByName(fileName + ".0.sig");
            Assert.Equal(fileName + ".0.sig", result.Name);
            Assert.Equal(1, result.Length);
        }

        [Fact]
        public void Should_get_SignaturInfos_by_file_name()
        {
            var tested = new StorageSignatureRepository(transactionalStorage, "test", configuration);

            transactionalStorage.Batch(accessor =>
            {
                accessor.AddSignature("test", 0, stream => stream.Write(new byte[] { 3 }, 0, 1));
                accessor.AddSignature("test", 1, stream => stream.Write(new byte[] { 3 }, 0, 1));
                accessor.AddSignature("test", 2, stream => stream.Write(new byte[] { 3 }, 0, 1));
            });

            var signatureInfos = tested.GetByFileName().ToList();
            Assert.Equal(3, signatureInfos.Count());
            foreach (var item in signatureInfos)
            {
                Assert.Equal(1, item.Length);
            }
        }
    }
}
